/* thermostat, first try */
/* $Id: thermo.c,v 1.5 2005-11-11 07:45:55 cm Exp $ */

#include <inttypes.h>
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <string.h>
#include <stdlib.h>

#include "thermo.h"

/* globals */
volatile time_struct time;

volatile uint16_t cur_temp = 0; /* temperature from last ADC conversion */
uint16_t temperature; /* filtered temperature */

uint16_t setpoint = SETPOINT;

int cursor_position[][2] =
  { { 21,1 }, /* dummy */
    { 0,0 }, /* DoW */
    { 5,0 }, /* hour */
    { 8,0 }, /* min */
    { 13,1 }}; /* setpoint */

int8_t position = 0;
int8_t cursor_x = 21;
int8_t cursor_y = 1;


timer_struct fixed_timers[] =
  { { 6, 30, B_WOCHENTAG, T_ACTION_RAISE },
    { 9, 00, B_WOCHENTAG, T_ACTION_LOWER },
    { 16, 00, B_WOCHENTAG, T_ACTION_RAISE },
    { 23, 00, B_WOCHENTAG, T_ACTION_LOWER },

    { 10, 00, B_WOCHENEND, T_ACTION_RAISE },
    { 23, 00, B_WOCHENEND, T_ACTION_LOWER } };

timer_struct except_timer;


#define SER_OUT_BUF 16
char ser_out_buf[SER_OUT_BUF];
char *ser_out_ptr = ser_out_buf;
char buf[SER_OUT_BUF];

volatile uint8_t uart_rcvd;

volatile uint8_t flags = 0xff;

/* key handling */
volatile uint8_t key_pressed = 0x0;
uint8_t key_state = 0x0;
uint8_t key_old = 0x0;

#define T_A 53/6
#define T_B 2271
#define TEMP(x) ((x)*T_A-T_B)


void main(void) {
  ioinit();
  for(;;) {
    if(flags & FLAG_TIME_TICK) {
      flags &= ~FLAG_TIME_TICK;
      tick_handler();
    }
    if(flags & FLAG_TIME_SECOND) {
      flags &= ~FLAG_TIME_SECOND;
      second_handler();
    }
    if(flags & FLAG_TIME_MINUTE) {
      flags &= ~FLAG_TIME_MINUTE;
      minute_handler();
    }
    if(key_pressed) {
      key_handler();
      update_display();
    }
  }
}

/* runs once per timer tick, i.e. TICKS_PER_SECOND times per second */
void tick_handler() {
  check_keys();
}

/* runs once a second, when the second is incremented */
void second_handler() {
  /* start new ADC conversion */
  ADCSRA |= _BV(ADSC);

  CTL_PORT = pid(setpoint,TEMP(cur_temp),0);

  update_display();
}

void update_display() {
  /* display time */

  lcd_gotoxy(LCD_LOC_TIME_X,LCD_LOC_TIME_Y);
  lcd_puts(downames[time.wday]);
  lcd_putc(' ');
  lcd_printnum(2,time.hour,0);
  lcd_putc(':');
  lcd_printnum(2,time.min,1);

  lcd_gotoxy(LCD_LOC_TEMP_X,LCD_LOC_TEMP_Y);
  lcd_printnum(2, temperature/100,0);
  lcd_putc('.');
  lcd_printnum(1, (temperature%100)/10,1);
  lcd_putc(DEGREE_CHAR);

  lcd_gotoxy(0,1);
#if 0
  lcd_printnum(3, (CTL_PORT-CTL_MIN)*100/(CTL_MAX-CTL_MIN), 0);
  lcd_putc('%');
#else
  lcd_printnum(3, CTL_PORT, 0);
  lcd_putc(' ');
#endif

  lcd_puts_P(" Soll ");
  lcd_printnum(2, setpoint/100,0);
  lcd_putc('.');
  lcd_printnum(1, (setpoint%100)/10,1);
  lcd_putc(DEGREE_CHAR);

  lcd_gotoxy(cursor_x,cursor_y);

  

  ser_puts(itoa(temperature,buf,10));
  ser_puts(" ");
  ser_puts(itoa(CTL_PORT,buf,10));
  ser_puts(" ");
  ser_puts(itoa(TEMP(cur_temp),buf,10));
  ser_puts("\r\n");
}

void minute_handler() {
  uint8_t action = T_ACTION_INVALID,
    i;

  for(i = 0; i < sizeof(fixed_timers)/sizeof(timer_struct); i++) {
    if(fixed_timers[i].hour == time.hour &&
       fixed_timers[i].min == time.min &&
       fixed_timers[i].wday & (1 << time.wday)) {
      action = fixed_timers[i].action;
    }
  }
  switch(action) {
  case T_ACTION_DEFAULT:
    setpoint = SETPOINT; break;
  case T_ACTION_LOWER:
    setpoint = SETPOINT_LOW; 
    (void)pid(setpoint,TEMP(cur_temp),1);
    break;
  case T_ACTION_RAISE:
    setpoint = SETPOINT;
    (void)pid(setpoint,TEMP(cur_temp),1);
    break;
  }
}
    

void key_handler() {

  if(key_pressed & KEY_LEFT) {
    position--;
    if(position < 0)
      position = 4;
  }
  if(key_pressed & KEY_RIGHT) {
    position++;
    if(position > 4)
      position = 0;
  }
  cursor_x = cursor_position[position][0];
  cursor_y = cursor_position[position][1];
  if(key_pressed & (KEY_LEFT|KEY_RIGHT))
    lcd_gotoxy(cursor_x,cursor_y);

  if(key_pressed & KEY_UP)
    switch(position) {
    case 1:
      time.wday++;
      if(time.wday > 6)
	time.wday = 0;
      break;
    case 2:
      time.hour++;
      if(time.hour > 23)
	time.hour = 0;
      break;
    case 3:
      time.min++;
      if(time.min > 59)
	time.min = 0;
      break;
    case 4:
      setpoint += 50;
      if(setpoint > SETPOINT_MAX)
	setpoint = SETPOINT_MAX;
      (void)pid(setpoint,TEMP(cur_temp),1);
      break;
    }

  if(key_pressed & KEY_DOWN)
    switch(position) {
    case 1:
      time.wday--;
      if(time.wday > 6)
	time.wday = 6;
      break;
    case 2:
      time.hour--;
      if(time.hour > 23)
	time.hour = 23;
      break;
    case 3:
      time.min--;
      if(time.min > 59)
	time.min = 59;
      break;
    case 4:
      setpoint -= 50;
      if(setpoint < SETPOINT_MIN)
	setpoint = SETPOINT_MIN;
      (void)pid(setpoint,TEMP(cur_temp),1);
      break;
    }


  /* empty */
  key_pressed = 0;
}

void ioinit() {
  /* external clock */
  /* runs Timer/Counter2 (8bit) */
  /* datasheet p116 ff */
  ASSR = _BV(AS2); /* source = TOSC1/2; write ASSR first! */
  TCNT2 = 0;
  TCCR2 = _BV(CS21); /* prescaler = /8, gives 16 interrupts/sec */
  while(ASSR & (_BV(TCN2UB)|_BV(TCR2UB))); /* settle */
  TIFR = (_BV(OCF2)|_BV(TOV2)); /* clear interrupt flags */
  timer_enable_int(_BV(TOIE2));

  /* Buttons */
  KEY_DDR &= ~KEY_MASK; /* set as inputs */
  KEY_PORT |= KEY_MASK; /* pullups enabled */

  /* ADC, datasheet p197 ff */
  /* general setup: enable, interrupt enable, prescaler /8 = 125kHz */
  ADCSRA = (_BV(ADEN)|_BV(ADIE)|_BV(ADPS1)|_BV(ADPS0));
  /* mux selection, reference */
  ADMUX = (_BV(REFS0) | /* AVcc with ext capacitor @AREF */
	   NTC_ADC); /* ADC0 as input */

  /* start ADC */
  ADCSRA |= _BV(ADSC);

  /* UART datasheet p130 ff */
  /* async dbl speed mode */
  UBRRH = SER_CLK>>8;
  UBRRL = SER_CLK&0xFF;
#ifdef SER_CLK_DBL
  UCSRA = _BV(U2X);
#endif
  UCSRB = (_BV(RXCIE)| /* interrupts on RX and TX */
	   _BV(TXEN)|_BV(RXEN)); /* enable TX/RX */
  UCSRC = (_BV(URSEL)|_BV(UCSZ1)|_BV(UCSZ0)); /* 8n1 */
  ser_puts("INIT\r\n");

  /* OC1A dims the LCD backlight via transistor */
  /* OC1B controls the heater via low-pass and transistor */
  /* timer 1, prescaler /1 */
  /* PWM phase correct 8bit */
  /* OC1A in "normal" mode, OC1B "inverse" mode */
  TCCR1A = _BV(COM1A1)|_BV(COM1B1)|_BV(COM1B0)|_BV(WGM10);
  TCCR1B = _BV(CS10); /* prescaler /1 */
  TCNT1 = 0;
  DDRB |= _BV(PB1)|_BV(PB2);
  OCR1A = LCD_BACKLIGHT_LEVEL;
  OCR1B = CTL_PWM_INIT;

  lcd_init(LCD_DISP_ON_BLINK);

  /* enable global interrupts */
  sei();

  /* wait for 1st ADC conversion and initialized IIR filter/PID */
  while(!cur_temp);
  (void)pid(SETPOINT,TEMP(cur_temp),1);
}

/* send string to serial */
int ser_puts(char *s) {
  size_t r;

  cli();
  r = strlcat(ser_out_buf, s, SER_OUT_BUF);
  sei();
  uart_ena_xmit;
  return (r >= SER_OUT_BUF); /* if >=, truncated */
}

/* display error on LCD and die */
void die(const char *msg) {
  lcd_clrscr();
  lcd_puts(msg);
  for(;;);
}


/* RTC interrupt */
SIGNAL(SIG_OVERFLOW2) {
  time.ticks++;
  flags |= FLAG_TIME_TICK;

  /* keep time */
  if(time.ticks >= TICKS_PER_SEC) {
    time.ticks = 0;
    time.sec++;
    flags |= FLAG_TIME_SECOND;
    if(time.sec > 59) {
      time.sec = 0;
      time.min++;
      flags |= FLAG_TIME_MINUTE;
      if(time.min > 59) {
	time.min = 0;
	time.hour++;
	flags |= FLAG_TIME_HOUR;
	if(time.hour > 23) {
	  time.hour = 0;
	  time.wday++;
	  flags |= FLAG_TIME_DAY;
	  if(time.wday > 6) {
	    time.wday = 0;
	    flags |= FLAG_TIME_WEEK;
	  }
	}
      }
    }
  }
}

SIGNAL(SIG_ADC) {
  cur_temp = ADCW;
}

SIGNAL(SIG_UART_RECV) {
  uint8_t status = UCSRA;

  if(status & _BV(FE)) {
    /* frame error */
  }
  if(status & _BV(DOR)) {
    /* data overrun error */
  }
  if(status & _BV(PE)) {
    /* parity error */
  }
  uart_rcvd = UDR;
}

SIGNAL(SIG_UART_DATA) {
  char c = *ser_out_ptr++;
  if(c)
    UDR = c;
  else {
    uart_dis_xmit;
    ser_out_ptr = ser_out_buf;
    *ser_out_buf = 0;
  }
}
  
/* key handling and debouncing; cf 
   http://www.mikrocontroller.net/articles/Entprellung
*/
void check_keys() {
  uint8_t iwr0, iwr1;

  iwr0 = key_old;
  key_old = KEY_PIN & KEY_MASK;
  iwr0 ^= key_old;
  key_old = ~key_old;
  iwr1 = key_state;
  key_state |= iwr0;
  iwr0 &= key_old;
  key_state ^= iwr0;
  iwr1 &= iwr0;
  key_pressed |= iwr1;
}

/* 
   $Log: thermo.c,v $
   Revision 1.5  2005-11-11 07:45:55  cm
   leave cursor on all the time

   Revision 1.4  2005/11/11 07:31:30  cm
   recalibrate NTC (resoldered)

   Revision 1.3  2005/11/09 08:42:40  cm
   initialize pid on setpoint change

   Revision 1.2  2005/10/26 22:48:23  cm
   working version; time + setpoint setting per keypad. pid suboptimal

   Revision 1.1.1.1  2005/10/26 16:18:13  cm
   import

*/
