#ifndef THERMO_H
#define THERMO_H

#define USE_BIGCHAR

#define SETPOINT 2150
#define SETPOINT_LOW 1800
#define SETPOINT_MIN 500
#define SETPOINT_MAX 2500

/* PWM duty cycle (max 255) for LCD LED backlight */
#define LCD_BACKLIGHT_LEVEL 25
/* PWM duty cycle (max 255) for heater modulation; */
/* a value of 0 should give ~0V output */
#define CTL_PWM_INIT 0
#define CTL_PORT OCR1B
#define CTL_OFF 95
#define CTL_MIN 110
#define CTL_MAX 210

/* PID controller parameter */
#define pid_P 1.152
/* was: 0.00024 @1sec, .00024/60 @1min? */
#define pid_I (0.00024*60)
#define pid_D 0.0

#define T_A 53/6
#define T_B 2241
#define TEMP(x) ((x)*T_A-T_B)



/* LCD cursor locations for various text */
#define LCD_LOC_TIME_X 0
#define LCD_LOC_TIME_Y 0
#define LCD_LOC_TIME_COLON_X 6

#define LCD_LOC_TEMP_X 10
#define LCD_LOC_TEMP_Y 0

/* LCD char code for "degree" character */
#define DEGREE_CHAR 0xdf
#define RIGHT_CHAR 0x7e
#define LEFT_CHAR 0x7f
#define BIGDOT_CHAR 0xa5

/* flag bits */
#define FLAG_TIME_TICK   1 /* time has advanced one tick */
#define FLAG_TIME_SECOND 2 /* time has advanced one second */
#define FLAG_TIME_MINUTE 4
#define FLAG_TIME_HOUR   8
#define FLAG_TIME_DAY   16
#define FLAG_TIME_WEEK  32
#define FLAG_UART_RCVD  64 /* new character from UART available */
extern volatile uint8_t flags;

extern uint16_t temperature;

extern uint16_t setpoint;


#define KEY_PORT PORTC
#define KEY_PIN PINC
#define KEY_DDR DDRC
#define KEY_1 (_BV(PC1))
#define KEY_2 (_BV(PC2))
#define KEY_3 (_BV(PC3))
#define KEY_4 (_BV(PC4))
#define KEY_MASK (KEY_1|KEY_2|KEY_3|KEY_4)
#define KEY_DOWN KEY_1
#define KEY_UP KEY_3
#define KEY_LEFT KEY_4
#define KEY_RIGHT KEY_2

#define NTC_ADC 0 /* ADC Mux port for NTC */

#define SER_BAUD 9600UL /* must be long for calculations */
#define SER_CLK ((XTAL)/(8*SER_BAUD)-1)
#define SER_CLK_DBL 1

#define TICKS_PER_SEC 16 /* change TCCR2 in ioinit() too! */

#define uart_ena_xmit UCSRB |= _BV(UDRIE);
#define uart_dis_xmit UCSRB &= ~ _BV(UDRIE);


#define T_ACTION_DEFAULT 0 /* set temp to default */
#define T_ACTION_LOWER 1   /* lower temp to night setting */
#define T_ACTION_RAISE 2   /* raise temp to day setting */
#define T_ACTION_INVALID 0xff

typedef struct {
  uint8_t hour;
  uint8_t min;
  uint8_t wday; /* bitfield! */
  uint8_t action;
} timer_struct;


/* temperature from last ADC conversion */
extern volatile uint16_t cur_temp;


void ioinit();

/*void die(char const *);*/

void busy_wait_ticks(uint8_t);

void usleep(uint8_t);

void tick_handler();
void second_handler();
void minute_handler();

int ser_puts(char *);

uint8_t pid(uint8_t);

void temp_filter(uint8_t);

void key_handler();

void update_display();


/* LCD defines */
#endif

/*
  $Log: thermo.h,v $
  Revision 1.5  2005-11-29 18:50:48  cm
  added bigchar support
  changed I for PID

  Revision 1.4  2005/11/15 23:47:26  cm
  changes

  Revision 1.3  2005/11/14 22:17:09  cm
  calibration
  pid 1/minute

  Revision 1.2  2005/10/26 22:48:23  cm
  working version; time + setpoint setting per keypad. pid suboptimal

  Revision 1.1.1.1  2005/10/26 16:18:13  cm
  import

*/
